@@ -4,7 +4,7 @@ use warnings;
 
 package RT::Extension::SLA;
 
-our $VERSION = '0.05';
+our $VERSION = '0.06';
 
 =head1 NAME
 
@@ -26,8 +26,26 @@ RT extension to implement automated due dates using service levels.
 
 =item make initdb (for the first time only)
 
+=item Base configuration
+
+In RT 3.8 and later, you must enable the plugin by adding RT::Extension::SLA
+to your @Plugins line (or create one) like:
+
+    Set(@Plugins,(qw(RT::Extension::SLA)));
+
 =back
 
+=head1 UPGRADING
+
+=head2 From versions prior to 0.06
+
+You need to run an upgrade step on your RT database so this extension continues
+to work.  Run the following from inside the source of this extension:
+
+    /opt/rt4/sbin/rt-setup-database --action insert --datafile etc/upgrade/0.06/content
+
+It will prompt you for your DBA password and should complete without error.
+
 =head1 CONFIGURATION
 
 Service level agreements of tickets is controlled by an SLA custom field (CF).
@@ -60,14 +78,14 @@ In this example I<Incident> is the name of the queue, and I<2h> is the name of
 the SLA which will be applied to this queue by default.
 
 Each service level can be described using several options:
-L<StartImmediately|/"StartImmediately (boolean, false)">,
+L<Starts|/"Starts (interval, first business minute)">,
 L<Resolve|/"Resolve and Response (interval, no defaults)">,
 L<Response|/"Resolve and Response (interval, no defaults)">,
 L<KeepInLoop|/"Keep in loop (interval, no defaults)">,
 L<OutOfHours|/"OutOfHours (struct, no default)">
 and L<ServiceBusinessHours|/"Configuring business hours">.
 
-=head2 StartImmediately (boolean, false)
+=head2 Starts (interval, first business minute)
 
 By default when a ticket is created Starts date is set to
 first business minute after time of creation. In other
@@ -77,8 +95,21 @@ be beginning of the next business day.
 
 However, if you provide 24/7 support then you most
 probably would be interested in Starts to be always equal
-to Created time. In this case you can set option
-StartImmediately to a true value.
+to Created time.
+
+Starts option can be used to adjust behaviour. Format
+of the option is the same as format for deadlines which
+described later in details. RealMinutes, BusinessMinutes
+options and OutOfHours modifiers can be used here like
+for any other deadline. For example:
+
+    'standard' => {
+        # give people 15 minutes
+        Starts   => { BusinessMinutes => 15  },
+    },
+
+You can still use old option StartImmediately to set
+Starts date equal to Created date.
 
 Example:
 
@@ -86,9 +117,12 @@ Example:
         StartImmediately => 1,
         Response => { RealMinutes => 30 },
     },
-    'standard' => {
-        StartImmediately => 0, # can be ommited as it's default
-        Response => { BusinessMinutes => 2*60 },
+
+But it's the same as:
+
+    '24/7' => {
+        Starts => { RealMinutes => 0 },
+        Response => { RealMinutes => 30 },
     },
 
 =head2 Resolve and Response (interval, no defaults)
@@ -219,7 +253,9 @@ In the above example Due is set to one hour after creation, reply
 of a non-requestor moves Due date two hours forward, requestors'
 replies move Due date to one hour and resolve deadine is 24 hours.
 
-=head2 OutOfHours (struct, no default)
+=head2 Modifying Agreements
+
+=head3 OutOfHours (struct, no default)
 
 Out of hours modifier. Adds more real or business minutes to resolve
 and/or reply options if event happens out of business hours, read also
@@ -243,6 +279,26 @@ hours, otherwise only one.
 Supporters have two additional hours in the morning to deal with bunch
 of requests that came into the system during the last night.
 
+=head3 IgnoreOnStatuses (array, no default)
+
+Allows you to ignore a deadline when ticket has certain status. Example:
+
+    'level x' => {
+        KeepInLoop => { BusinessMinutes => 60, IgnoreOnStatuses => ['stalled'] },
+    },
+
+In above example KeepInLoop deadline is ignored if ticket is stalled.
+
+B<NOTE>: When a ticket goes from an ignored status to a normal status, the new
+Due date is calculated from the last action (reply, SLA change, etc) which fits
+the SLA type (Response, Starts, KeepInLoop, etc).  This means if a ticket in
+the above example flips from stalled to open without a reply, the ticket will
+probably be overdue.  In most cases this shouldn't be a problem since moving
+out of stalled-like statuses is often the result of RT's auto-open on reply
+scrip, therefore ensuring there's a new reply to calculate Due from.  The
+overall effect is that ignored statuses don't let the Due date drift
+arbitrarily, which could wreak havoc on your SLA performance.
+
 =head2 Configuring business hours
 
 In the config you can set one or more work schedules. Use the following
@@ -321,10 +377,27 @@ sub BusinessHours {
 
 sub Agreement {
     my $self = shift;
-    my %args = ( Level => undef, Type => 'Response', Time => undef, @_ );
+    my %args = (
+        Level => undef,
+        Type => 'Response',
+        Time => undef,
+        Ticket => undef,
+        Queue  => undef,
+        @_
+    );
 
     my $meta = $RT::ServiceAgreements{'Levels'}{ $args{'Level'} };
     return undef unless $meta;
+
+    if ( exists $meta->{'StartImmediately'} || !defined $meta->{'Starts'} ) {
+        $meta->{'Starts'} = {
+            delete $meta->{'StartImmediately'}
+                ? ( )
+                : ( BusinessMinutes => 0 )
+            ,
+        };
+    }
+
     return undef unless $meta->{ $args{'Type'} };
 
     my %res;
@@ -337,55 +410,78 @@ sub Agreement {
         return undef;
     }
 
-    if ( defined $meta->{'StartImmediately'} ) {
-        $res{'StartImmediately'} = $meta->{'StartImmediately'};
+    if ( $args{'Ticket'} && $res{'IgnoreOnStatuses'} ) {
+        my $status = $args{'Ticket'}->Status;
+        return undef if grep $_ eq $status, @{$res{'IgnoreOnStatuses'}};
     }
 
-    if ( $args{'Time'} and my $tmp = $meta->{'OutOfHours'}{ $args{'Type'} } ) {
-        my $bhours = $self->BusinessHours( $meta->{'BusinessHours'} );
-        if ( $bhours->first_after( $args{'Time'} ) != $args{'Time'} ) {
-            foreach ( qw(RealMinutes BusinessMinutes) ) {
-                next unless $tmp->{ $_ };
-                $res{ $_ } ||= 0;
-                $res{ $_ } += $tmp->{ $_ };
-            }
-        }
+    $res{'OutOfHours'} = $meta->{'OutOfHours'}{ $args{'Type'} };
+
+    $args{'Queue'} ||= $args{'Ticket'}->QueueObj if $args{'Ticket'};
+    if ( $args{'Queue'} && ref $RT::ServiceAgreements{'QueueDefault'}{ $args{'Queue'}->Name } ) {
+        $res{'Timezone'} = $RT::ServiceAgreements{'QueueDefault'}{ $args{'Queue'}->Name }{'Timezone'};
     }
+    $res{'Timezone'} ||= $meta->{'Timezone'} || $RT::Timezone;
+
+    $res{'BusinessHours'} = $meta->{'BusinessHours'};
 
     return \%res;
 }
 
 sub Due {
     my $self = shift;
-    my %args = ( Level => undef, Type => undef, Time => undef, @_ );
-
-    my $agreement = $self->Agreement( %args );
-    return undef unless $agreement;
-
-    my $meta = $RT::ServiceAgreements{'Levels'}{ $args{'Level'} };
-
-    my $res = $args{'Time'};
-    if ( defined $agreement->{'BusinessMinutes'} ) {
-        my $bhours = $self->BusinessHours( $meta->{'BusinessHours'} );
-        $res = $bhours->add_seconds( $res, 60 * $agreement->{'BusinessMinutes'} );
-    }
-    $res += 60 * $agreement->{'RealMinutes'}
-        if defined $agreement->{'RealMinutes'};
-
-    return $res;
+    return $self->CalculateTime( @_ );
 }
 
 sub Starts {
     my $self = shift;
-    my %args = ( Level => undef, Time => undef, @_ );
+    return $self->CalculateTime( @_, Type => 'Starts' );
+}
 
-    my $meta = $RT::ServiceAgreements{'Levels'}{ $args{'Level'} };
-    return undef unless $meta;
+sub CalculateTime {
+    my $self = shift;
+    my %args = (@_);
+    my $agreement = $args{'Agreement'} || $self->Agreement( @_ );
+    return undef unless $agreement and ref $agreement eq 'HASH';
 
-    return $args{'Time'} if $meta->{'StartImmediately'};
+    my $res = $args{'Time'};
 
-    my $bhours = $self->BusinessHours( $meta->{'BusinessHours'} );
-    return $bhours->first_after( $args{'Time'} );
+    my $ok = eval {
+        local $ENV{'TZ'} = $ENV{'TZ'};
+        if ( $agreement->{'Timezone'} && $agreement->{'Timezone'} ne ($ENV{'TZ'}||'') ) {
+            $ENV{'TZ'} = $agreement->{'Timezone'};
+            require POSIX; POSIX::tzset();
+        }
+
+        my $bhours = $self->BusinessHours( $agreement->{'BusinessHours'} );
+
+        if ( $agreement->{'OutOfHours'} && $bhours->first_after( $res ) != $res ) {
+            foreach ( qw(RealMinutes BusinessMinutes) ) {
+                next unless my $mod = $agreement->{'OutOfHours'}{ $_ };
+                ($agreement->{ $_ } ||= 0) += $mod;
+            }
+        }
+
+        if ( defined $agreement->{'BusinessMinutes'} ) {
+            if ( $agreement->{'BusinessMinutes'} ) {
+                $res = $bhours->add_seconds(
+                    $res, 60 * $agreement->{'BusinessMinutes'},
+                );
+            }
+            else {
+                $res = $bhours->first_after( $res );
+            }
+        }
+        $res += 60 * $agreement->{'RealMinutes'}
+            if defined $agreement->{'RealMinutes'};
+        1;
+    };
+
+    POSIX::tzset() if $agreement->{'Timezone'}
+        && $agreement->{'Timezone'} ne ($ENV{'TZ'}||'');
+    die $@ unless $ok;
+
+    return $res;
 }
 
 sub GetCustomField {
@@ -420,8 +516,10 @@ sub GetDefaultServiceLevel {
         else {
             return $args{'Queue'}->SLA if $args{'Queue'}->SLA;
         }
-        if ( $RT::ServiceAgreements{'QueueDefault'} && $RT::ServiceAgreements{'QueueDefault'}{ $args{'Queue'}->Name } ) {
-            return $RT::ServiceAgreements{'QueueDefault'}{ $args{'Queue'}->Name };
+
+        if ( my $info = $RT::ServiceAgreements{'QueueDefault'}{ $args{'Queue'}->Name } ) {
+            return $info unless ref $info;
+            return $info->{'Level'} || $RT::ServiceAgreements{'Default'};
         }
     }
     return $RT::ServiceAgreements{'Default'};
