@@ -2,7 +2,7 @@ use strict;
 use warnings;
 
 ### after: use lib qw(@RT_LIB_PATH@);
-use lib qw(/opt/rt3/local/lib /opt/rt3/lib);
+use lib qw(/home/tom/work/bps/rt/local/lib /home/tom/bps/rt/lib);
 
 package RT::Extension::SLA::Test;
 
